
#include "x.h"
#include "xt.h"
#include "xpex.h"
#include <math.h>
#include "book_utils.h"
#include "externs.h"
#include "defs.h"

XtAppContext  app;
Display      *dpy;
Window        window, windowWindow;
Drawable      drawable;
Colormap      colormap;
Multibuffer   buffers[2];
Pixmap        pixmap;
XWindowAttributes    xwattrs;
GC            gc;
Atom          delete_window_atom;

PEXRenderer           renderer;
PEXRendererAttributes rattrs;

PEXStructure top_struct, view_struct, sphere;

Widget toplevel, canvas, canvasWindow;

PEXColorSpecifier colorSpec;

PEXTableIndex        listON[4] = { 1, 2 };
/* PEXTableIndex     listOFF[4] = { 3, 4 }; */
PEXTableIndex        listOFF[4];
PEXLightEntry        lights[4];
PEXDepthCueEntry        depth_cue_rep;

/* The viewing parameters. */
PEXCoord         view_ref_pt;
PEXVector        view_up_vec;
PEXVector        view_plane_normal;
PEXCoord2D       view_window[2];
PEXNPCSubVolume  viewport;
double           view_plane, front_plane, back_plane;
PEXCoord         proj_ref_pt;
int              perspective;
PEXViewEntry     view;

int using_drawable = BUFFER;

main( argc, argv )
int   argc;
char *argv[];
{
    XVisualInfo            vis_info;
    XStandardColormap        cmap_info;
    PEXColorApproxEntry        capx_info;
    XSetWindowAttributes    wattrs;
    PEXExtensionInfo        *pexinfo;
    char            err[80];
    extern int PrintDefaultError();
    int                  type;
    XColor background;
    XWindowAttributes xwattrs;

    unsigned long   pcMask[3];
    PEXPCAttributes	pcattrs;

    initialize_view();

    if( argc > 1 ) {
    float stof();
    view_ref_pt.x = stof( argv[2] );
    view_ref_pt.y = stof( argv[3] );
    view_ref_pt.z = stof( argv[4] );
    view_up_vec.x = stof( argv[6] );
    view_up_vec.y = stof( argv[7] );
    view_up_vec.z = stof( argv[8] );
    view_plane_normal.x = stof( argv[10] );
    view_plane_normal.y = stof( argv[11] );
    view_plane_normal.z = stof( argv[12] );
    proj_ref_pt.x = stof( argv[14] );
    proj_ref_pt.y = stof( argv[15] );
    proj_ref_pt.z = stof( argv[16] );
    back_plane = stof( argv[18] );
    view_plane = stof( argv[20] );
    front_plane = stof( argv[22] );
    perspective = stof( argv[24] );
    view_window[0].x = stof( argv[27] );
    view_window[0].y = stof( argv[28] );
    view_window[1].x = stof( argv[30] );
    view_window[1].y = stof( argv[31] );
    viewport.min.x = stof( argv[34] );
    viewport.min.y = stof( argv[35] );
    viewport.min.z = stof( argv[36] );
    viewport.max.x = stof( argv[38] );
    viewport.max.y = stof( argv[39] );
    viewport.max.z = stof( argv[40] );
    view.clip_flags = stof( argv[43] );
    view.clip_limits.min.x = stof( argv[46] );
    view.clip_limits.min.y = stof( argv[47] );
    view.clip_limits.min.z = stof( argv[48] );
    view.clip_limits.max.x = stof( argv[50] );
    view.clip_limits.max.y = stof( argv[51] );
    view.clip_limits.max.z = stof( argv[52] );
    }

    /* Initialize the toolkit. */
    toplevel = XtVaAppInitialize( &app, "XViewEditor",
        (XrmOptionDescList)NULL, 0,
        (int *)&argc, (String *)argv,
        (String *)NULL, NULL );

    /* See what display is being used. */
    dpy = XtDisplay( toplevel );
    XSetErrorHandler( PrintDefaultError );

    /* See if a PEX extension exists and initialize PEXlib. */
    if ( PEXInitialize( dpy, &pexinfo, 80, err ) != 0 ) {
        fprintf( stderr, "No PEX extension.\n" );
        fprintf( stderr, "%s\n", err );
        exit(1);
    }

    /* Determine the best visual to use. */
    ora_find_best_visual( dpy, &vis_info );

    if( vis_info.depth == 24 )
        ora_create_colormap( dpy, &vis_info, &cmap_info, &capx_info );
    else {
        /* Get a standard colormap for the visual. */
        if ( !ora_get_standard_colormap( dpy, &vis_info, &cmap_info ) ) {
            fprintf ( stderr, "Cannot find a standard colormap\n" );
            exit(1);
        }
        ora_set_stdcmap_approx( &vis_info, &cmap_info, &capx_info );
    }

    if( !(strcmp( pexinfo->vendor_name, "International Business Machines" ) ) )
        XtVaSetValues( toplevel,
            XtNvisual, vis_info.visual,
            XtNdepth, vis_info.depth,
            XtNcolormap, cmap_info.colormap,
            NULL );

    create_gui( vis_info.depth, pexinfo->vendor_name );
    XtRealizeWidget( toplevel );

    if( !(strcmp( pexinfo->vendor_name, "International Business Machines" )) ) {
        window = XtWindow( canvas );
        windowWindow = XtWindow( canvasWindow );
        XGetWindowAttributes( dpy, window, &xwattrs );
        gc = DefaultGC( dpy, DefaultScreen(dpy) );
        XSelectInput( dpy, windowWindow, ExposureMask | Button1MotionMask | ButtonPressMask | ButtonReleaseMask );
    }
    else {
        background.pixel = 0;
        background.flags = DoRed | DoGreen | DoBlue;
        window = ora_create_window( dpy, XtWindow(canvas),
            &vis_info, &cmap_info,
            "210x210", &background );
        windowWindow = ora_create_window( dpy, XtWindow(canvasWindow),
            &vis_info, &cmap_info,
            "210x210", &background );
        XGetWindowAttributes( dpy, window, &xwattrs );
        gc = DefaultGC( dpy, DefaultScreen(dpy) );
        XSelectInput( dpy, windowWindow, ExposureMask | 
            ButtonPressMask | Button1MotionMask | ButtonReleaseMask );
    }

    if( PEXUtDBConfig( dpy, window, True, &background, False, &drawable ) ) {
        fprintf( stderr, "No Double Buffering Support\n" );
        drawable = window;
    }

    /* Create the PEX renderer. */
    if ( IMMED_MODE_SPT( pexinfo ) ) {
        renderer = ora_setup_renderer( dpy, drawable, &capx_info, &rattrs );
    } else {
        fprintf( stderr, "Immediate mode not supported.\n" );
        exit(1);
    }

    /* Set the renderer's Subvolume. */
    rattrs.npc_subvolume.min.x = 0.; rattrs.npc_subvolume.max.x = 1.;
    rattrs.npc_subvolume.min.y = 0.; rattrs.npc_subvolume.max.y = 1.;
    rattrs.npc_subvolume.min.z = 0.; rattrs.npc_subvolume.max.z = 1.;
    /* Set the renderer's Viewport. */
    rattrs.viewport.min.x = 0;       rattrs.viewport.max.x = xwattrs.width;
    rattrs.viewport.min.y = 0;       rattrs.viewport.max.y = xwattrs.height;
    rattrs.viewport.min.z = 0.;      rattrs.viewport.max.z = 1.;
    rattrs.viewport.use_drawable = True;
    PEXChangeRenderer( dpy, renderer,
        PEXRANPCSubVolume | PEXRAViewport, &rattrs );

    pcMask[0] = pcMask[1] = pcMask[2] = 0;
    pcattrs.marker_type          = PEXMarkerAsterisk;
    PEXSetPCAttributeMask( pcMask, PEXPCMarkerType );
    pcattrs.marker_scale         = 2.;
    PEXSetPCAttributeMask( pcMask, PEXPCMarkerScale );
    pcattrs.marker_color.type    = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 0.0, pcattrs.marker_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCMarkerColor );
    pcattrs.distinguish          = True;
    PEXSetPCAttributeMask( pcMask, PEXPCDistinguishFlag );
    pcattrs.interior_style       = PEXInteriorStyleSolid;
    PEXSetPCAttributeMask( pcMask, PEXPCInteriorStyle );
    pcattrs.bf_interior_style    = PEXInteriorStyleSolid;
    PEXSetPCAttributeMask( pcMask, PEXPCBFInteriorStyle );
    pcattrs.reflection_model     = PEXReflectionSpecular;
    PEXSetPCAttributeMask( pcMask, PEXPCReflectionModel );
    pcattrs.bf_reflection_model  = PEXReflectionSpecular;
    PEXSetPCAttributeMask( pcMask, PEXPCBFReflectionModel );
    pcattrs.light_state.count    = 2;
    pcattrs.light_state.indices  = listON;
    PEXSetPCAttributeMask( pcMask, PEXPCLightState );
    pcattrs.surface_interp       = PEXSurfaceInterpColor;
    PEXSetPCAttributeMask( pcMask, PEXPCSurfaceInterp );
    pcattrs.bf_surface_interp    = PEXSurfaceInterpColor;
    PEXSetPCAttributeMask( pcMask, PEXPCBFSurfaceInterp );
    pcattrs.depth_cue_index      = 1;
    PEXSetPCAttributeMask( pcMask, PEXPCDepthCueIndex );
    pcattrs.reflection_attr.ambient = 0.4;
    pcattrs.reflection_attr.diffuse = 0.4;
    pcattrs.reflection_attr.specular = 0.5;
    pcattrs.reflection_attr.specular_conc = 3.0;
    pcattrs.reflection_attr.transmission = 0.0;
    pcattrs.reflection_attr.specular_color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, pcattrs.reflection_attr.specular_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCReflectionAttr );
    pcattrs.bf_reflection_attr   = pcattrs.reflection_attr;
    PEXSetPCAttributeMask( pcMask, PEXPCBFReflectionAttr );
    pcattrs.surface_color.type            = PEXColorTypeRGB;
    SET_COLOR( 1, 0.2, 0.5, pcattrs.surface_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCSurfaceColor );
    pcattrs.bf_surface_color.type            = PEXColorTypeRGB;
    SET_COLOR( 0.5, 0.5, 0.5, pcattrs.bf_surface_color.value );
    PEXSetPCAttributeMask( pcMask, PEXPCBFSurfaceColor );
    rattrs.pipeline_context = 
    PEXCreatePipelineContext( dpy, pcMask, &pcattrs );
    PEXChangeRenderer( dpy, renderer, PEXRAPipelineContext, &rattrs );

    build_sphere();

    top_struct = PEXCreateStructure( dpy );
    view_struct = PEXCreateStructure( dpy );

    PEXSetTableEntries( dpy, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetViewIndex( dpy, top_struct, PEXOCStore, 1 );
    PEXSetViewIndex( dpy, view_struct, PEXOCStore, 1 );
    PEXLabel( dpy, top_struct, PEXOCStore, OBJECT_LABEL );
    PEXExecuteStructure( dpy, top_struct, PEXOCStore, sphere );

    initialize_light_table();
    initialize_depth_cue_table();
{
    PEXCoord *points_vpn, tmp_vpn;
    PEXCoord *points_vup, tmp_vup;
    PEXColor  color;

    PEXLabel( dpy, view_struct, PEXOCStore, VRP_LABEL );
    PEXMarkers( dpy, view_struct, PEXOCStore, 1, &view_ref_pt );

    PEXLabel( dpy, view_struct, PEXOCStore, PRP_LABEL );
    PEXMarkers( dpy, view_struct, PEXOCStore, 1, &proj_ref_pt );

    SET_COLOR( 0, 1, 0, color );
    PEXSetLineColor( dpy, view_struct, PEXOCStore, PEXColorTypeRGB, &color );

    points_vup = (PEXCoord *)malloc( 2 * sizeof(PEXCoord) );
    points_vup[0] = view_ref_pt;
    tmp_vup.x = view_up_vec.x;
    tmp_vup.y = view_up_vec.y;
    tmp_vup.z = view_up_vec.z;
    points_vup[1] = tmp_vup;
    PEXLabel( dpy, view_struct, PEXOCStore, VUP_LABEL );
    PEXPolyline( dpy, view_struct, PEXOCStore, 2, points_vup );

    SET_COLOR( 0, 0, 1, color );
    PEXSetLineColor( dpy, view_struct, PEXOCStore, PEXColorTypeRGB, &color );

    points_vpn = (PEXCoord *)malloc( 2 * sizeof(PEXCoord) );
    points_vpn[0] = view_ref_pt;
    tmp_vpn.x = view_plane_normal.x;
    tmp_vpn.y = view_plane_normal.y;
    tmp_vpn.z = view_plane_normal.z;
    points_vpn[1] = tmp_vpn;
    PEXLabel( dpy, view_struct, PEXOCStore, VPN_LABEL );
    PEXPolyline( dpy, view_struct, PEXOCStore, 2, points_vpn );
}
    for (;;) {
        XEvent event;
        XtAppNextEvent(app, &event);
        if( (event.type == Expose) && 
            (event.xexpose.window == window) &&
            (event.xexpose.count == 0) ) {
            redraw();
        }
        else if( (event.type == Expose) && 
            (event.xexpose.window == windowWindow) &&
            (event.xexpose.count == 0) ) {
            redrawWindow();
        }
        else
            XtDispatchEvent(&event);
    }
}


float stof( buf )
char *buf;
{
    float f;
    sscanf( buf, "%f", &f );
    return f;
}


void
redrawWindow()
{
    XWindowAttributes xwattrs;
    int width, height;

    XGetWindowAttributes( dpy, windowWindow, &xwattrs );
    width = xwattrs.width;
    height = xwattrs.height;

    XSetForeground( dpy, gc, 0 );
    XDrawRectangle( dpy, windowWindow, gc,
        (int)(((float)width/2.) + view_window[0].x * (float)width / 10.),
        width-(int)(((float)width/2.) + view_window[1].y * (float)width / 10.),
        (int)((view_window[1].x-view_window[0].x) * (float)width / 10.),
        (int)((view_window[1].y-view_window[0].y) * (float)width / 10.) );
}


void
redraw()
{
    rattrs.hlhsr_mode = PEXHLHSRZBuffer;
    rattrs.clear_image = True;
    rattrs.clear_z = True;
    PEXChangeRenderer( dpy, renderer,
        PEXRAHLHSRMode | PEXRAClearImage | PEXRAClearZ, &rattrs );
    PEXRenderNetwork( dpy, drawable, renderer, top_struct );
    rattrs.hlhsr_mode = PEXHLHSROff;
    rattrs.clear_image = False;
    rattrs.clear_z = False;
    PEXChangeRenderer( dpy, renderer,
        PEXRAHLHSRMode | PEXRAClearImage | PEXRAClearZ, &rattrs );
    PEXRenderNetwork( dpy, drawable, renderer, view_struct );

    drawable = PEXUtDBSwap( dpy, window, False );
}


#define NORMALIZE( _v ) \
{ \
    double    mag; \
\
    mag = 1.0 / sqrt((_v).x * (_v).x \
          + (_v).y * (_v).y + (_v).z * (_v).z); \
    (_v).x *= mag, (_v).y *= mag, (_v).z *= mag; \
}

static void
initialize_light_table()
{
    lights[AMBIENT].type = PEXLightAmbient;
    lights[AMBIENT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[AMBIENT].color.value );
    PEXSetTableEntries( dpy, rattrs.light_table, 1, 1,
        PEXLUTLight, &lights[AMBIENT] );

    lights[DIRECTIONAL].type = PEXLightWCVector;
    lights[DIRECTIONAL].direction.x = -1;
    lights[DIRECTIONAL].direction.y =  0;
    lights[DIRECTIONAL].direction.z =  0;
    NORMALIZE(lights[DIRECTIONAL].direction);
    lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[DIRECTIONAL].color.value );
    PEXSetTableEntries( dpy, rattrs.light_table, 2, 1,
        PEXLUTLight, &lights[DIRECTIONAL] );
#if( 0 ) 
    lights[POSITIONAL].type = PEXLightWCPoint;
    lights[POSITIONAL].point.x = 0.5;
    lights[POSITIONAL].point.y = 2.0;
    lights[POSITIONAL].point.z = 0.5;
    lights[POSITIONAL].attenuation1 = 0;
    lights[POSITIONAL].attenuation2 = 1;
    lights[POSITIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[POSITIONAL].color.value );
    PEXSetTableEntries( dpy, rattrs.light_table, 3, 1,
        PEXLUTLight, &lights[POSITIONAL] );

    lights[SPOT].type = PEXLightWCSpot;
    lights[SPOT].point.x = 0.5;
    lights[SPOT].point.y = 2.0;
    lights[SPOT].point.z = 0.5;
    lights[SPOT].direction.x = 0;
    lights[SPOT].direction.y = -1;
    lights[SPOT].direction.z = 0;
    NORMALIZE(lights[SPOT].direction);
    lights[SPOT].attenuation1 = 0;
    lights[SPOT].attenuation2 = 1;
    lights[SPOT].concentration = 2;
    lights[SPOT].spread_angle = M_PI/2;
    lights[SPOT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1, 1, 1, lights[SPOT].color.value );
    PEXSetTableEntries( dpy, rattrs.light_table, 4, 1,
        PEXLUTLight, &lights[SPOT] );
#endif		/*  */
}

void
initialize_depth_cue_table()
{
    depth_cue_rep.mode = PEXOff;
    depth_cue_rep.front_plane = 1;
    depth_cue_rep.back_plane = 0;
    depth_cue_rep.front_scaling = 1;
    depth_cue_rep.back_scaling = 0.5;
    depth_cue_rep.color.type = PEXColorTypeRGB;
    SET_COLOR( 0, 0, 0, depth_cue_rep.color.value );
    PEXSetTableEntries( dpy, rattrs.depth_cue_table, 1, 1,
        PEXLUTDepthCue, &depth_cue_rep );
}

void
initialize_view()
{
    int error;

    view_ref_pt.x = 0.5;
    view_ref_pt.y = 0.5;
    view_ref_pt.z = 0.5;
    view_plane_normal.x = 1.;
    view_plane_normal.y = 1.2;
    view_plane_normal.z = 1.;
    view_up_vec.x = 0.;
    view_up_vec.y = 1.;
    view_up_vec.z = 0.;

    /* Compute the view orientation matrix. */
    error = PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    /* Set the view mapping parameters. */
    perspective = True;
    proj_ref_pt.x = 0; 
    proj_ref_pt.y = 0; 
    proj_ref_pt.z = 10.0;
    view_window[0].x = -1.5; 
    view_window[1].x =  1.5;
    view_window[0].y = -1.5; 
    view_window[1].y =  1.5;
    back_plane = -2.0;
    front_plane = 2.0; 
    view_plane = 0.0;
    viewport.min.x = 0.0;    viewport.max.x = 1.0;
    viewport.min.y = 0.0;    viewport.max.y = 1.0;
    viewport.min.z = 0.0;    viewport.max.z = 1.0;
    error = PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_pt, view_plane, back_plane,
        front_plane, view.mapping );

    /* Set the view representation. */
    view.clip_flags = PEXClippingAll;
    view.clip_limits = viewport;
}


#define NUM_LONG_FACETS   32
#define NUM_LONG_POINTS   (NUM_LONG_FACETS + 1)
#define NUM_LAT_FACETS    16
#define NUM_LAT_POINTS    (NUM_LAT_FACETS + 1)
#define RADIUS            1.00

void
build_sphere()
{
    PEXVertexNormal     vertex_data[NUM_LONG_POINTS * NUM_LAT_POINTS];
    PEXArrayOfVertex    vdata;
    PEXArrayOfFacetData facet_data;
    PEXVector           facet_normals[NUM_LONG_FACETS * NUM_LAT_FACETS];
    unsigned int        facet_data_flag, vertex_data_flag;
    double              theta = 0, phi = 0;
    register int        i, j, index;

    facet_data_flag = PEXGANormal;
    vertex_data_flag = PEXGANormal;

    for ( i = 0; i < NUM_LAT_POINTS; i++ ) {
        for ( j = 0; j < NUM_LONG_FACETS; j++ ) {
            index = i * NUM_LONG_POINTS + j;
            vertex_data[index].point.x = RADIUS * (1 + sin(phi) * cos(theta));
            vertex_data[index].point.y = RADIUS * (1 + cos(phi));
            vertex_data[index].point.z = RADIUS * (1 + sin(phi) * sin(theta));
            vertex_data[index].normal.x = sin(phi) * cos(theta);
            vertex_data[index].normal.y = cos(phi);
            vertex_data[index].normal.z = sin(phi) * sin(theta);
            NORMALIZE(vertex_data[index].normal);
            theta += (2 * M_PI) / NUM_LONG_FACETS;
        }
        phi += M_PI / NUM_LAT_FACETS;

        /* Close the circle. */
        index = i * NUM_LONG_POINTS + NUM_LONG_FACETS;
        vertex_data[index].point = vertex_data[index - NUM_LONG_FACETS].point;
        vertex_data[index].normal = vertex_data[index - NUM_LONG_FACETS].normal;
    }

    theta = M_PI / NUM_LONG_FACETS;
    phi = 0.5 * M_PI / NUM_LAT_FACETS;
    index = 0;
    for ( i = 0; i < NUM_LAT_FACETS; i++ ) {
        for ( j = 0; j < NUM_LONG_FACETS; j++ ) {
            facet_normals[index].x = sin(phi) * cos(theta);
            facet_normals[index].y = cos(phi);
            facet_normals[index].z = sin(phi) * sin(theta);
            ++index;
            theta += (2 * M_PI) / NUM_LONG_FACETS;
        }
        phi += M_PI / NUM_LAT_FACETS;
    }

    sphere = PEXCreateStructure( dpy );
    vdata.normal = vertex_data;
    facet_data.normal = facet_normals;
    PEXQuadrilateralMesh( dpy, sphere, PEXOCStore, PEXShapeUnknown,
        facet_data_flag, vertex_data_flag,
        PEXColorTypeRGB, facet_data,
        NUM_LONG_POINTS, NUM_LAT_POINTS, vdata );
}
