
#include "x.h"
#include "xt.h"
#ifndef USE_ATHENA
#include "xm.h"
#endif
#include "xpex.h"
#include "unix.h"
#include "defs.h"
#include "off.h"
#include "externs.h"
#include "dialog.h"
#include <sys/stat.h>

OFFObjDesc *Obj;        /* Object data structure to fill out */
char       *FileName;   /* File to be opened and read */
char      **environment;
int         numObjects;

extern XtAppContext app_context;

extern Widget toplevel, w_visual,
       w_wireframe, w_cull, w_edge, w_shadow, w_vanish,
       w_material_predef, w_video, w_hardcopy,
       w_reset, w_color, w_material, w_light, w_view, w_depthCue,
       w_flybycr, w_flybyex, w_animcr, w_animex, w_morph, w_spin, w_melt;

Dialog dialog_s = (Dialog)NULL,
       dialog_e = (Dialog)NULL,
       dialog_g = (Dialog)NULL,
       dialog_a = (Dialog)NULL;

int object_defined = 0;
int file_defined = 0;
int no_axis = 0;
int no_darea = 1;
int no_move = 0;
int no_background = 0;
int using_drawable = BUFFER;
int smth_changed = 0;
int verbose = 0;
int sofas = 0;
int track = 0;
int left_key_press = 0;


void main( argc, argv, envp )
int argc;
char *argv[];
char *envp[];
{
    extern void PrintSignalError();
    extern void AppMainLoop();
    environment = envp;

    if( strlen( getenv( "OBJ_PATH" ) ) == 0 ) {
        fprintf( stderr, "Environment variable OBJ_PATH, undefined\n\n" );
        exit( -1 );
    }

    parseXtAppArgs( argc, argv );
#ifndef NO_SIGNAL
    signal( SIGBUS,  PrintSignalError );	/* Bus Error */
    signal( SIGSEGV, PrintSignalError );	/* Segmentation Violation */
    signal( SIGILL,  PrintSignalError );	/* Illegal instruction */
    signal( SIGFPE,  SIG_IGN );				/* Ignore divide-by-zero traps */
#endif
    initXtAppTop( argc, argv );

    createControlPanel();

    if( object_defined ) {
        if( (Obj = (OFFObjDesc *)malloc( sizeof(OFFObjDesc) )) ==
            (OFFObjDesc *)0 ) {
            fprintf( stderr, "Cant allocate OFFObjDesc\n" );
            exit(1);
        }
        OFFReadObj( Obj );
        initPex( Obj );
    }
    else if( !file_defined ) {
        XtSetSensitive( w_reset, False );
        XtSetSensitive( w_visual, False );
        XtSetSensitive( w_wireframe, False );
        XtSetSensitive( w_cull, False );
        XtSetSensitive( w_edge, False );
        XtSetSensitive( w_shadow, False );
        XtSetSensitive( w_morph, False );
        XtSetSensitive( w_melt, False );
        XtSetSensitive( w_vanish, False );
        XtSetSensitive( w_color, False );
        XtSetSensitive( w_material, False );
        XtSetSensitive( w_view, False );
        XtSetSensitive( w_depthCue, False );
        XtSetSensitive( w_light, False );
        XtSetSensitive( w_material_predef, False );
        XtSetSensitive( w_flybycr, False );
        XtSetSensitive( w_flybyex, False );
        XtSetSensitive( w_animcr, False );
        XtSetSensitive( w_animex, False );
        XtSetSensitive( w_video, False );
        XtSetSensitive( w_hardcopy, False );
    }

    if( !no_darea ) XtAppMainLoop( app_context );
    else            AppMainLoop();
}


void parseXtAppArgs( argc, argv )
unsigned int argc;
char *argv[];
{
    int i;
    for( i=1; i<argc; i++ ) {
        if( (strncmp(argv[i], "-d", 2) == 0) || 
            (strcmp(argv[i], "-drawable") == 0) ) {
            if (++i >= argc) usage( argv[0] );
            if( strncmp( "buffer", argv[i], 6 ) == 0 )
                using_drawable = BUFFER;
            else if( strncmp( "pixmap", argv[i], 6 ) == 0 )
                using_drawable = PIXMAP;
            else if( strncmp( "window", argv[i], 6 ) == 0 )
                using_drawable = WINDOW;
            continue;
        }
        else if( (strncmp(argv[i], "-o", 2) == 0) || 
            (strcmp(argv[i], "-object") == 0) ) {
            struct stat s_buf;
            char *path;
            if (++i >= argc) usage( argv[0] );
            if( (FileName = (char *)malloc( strlen( argv[i] ) + 6 )) ==
                (char *)0 ) {
                fprintf( stderr, "Cant allocate char\n" );
                exit(1);
            }
            strcat( strcpy( FileName, argv[i] ), ".aoff" );
            if( (path = (char *)malloc( strlen(getenv("OBJ_PATH")) +
                                        strlen(FileName) +2 )) ==
                (char *)0 ) {
                fprintf( stderr, "Cant allocate char\n" );
                exit(1);
            }
            strcpy( path, getenv("OBJ_PATH") );
            strcat( path, "/" );
            strcat( path, FileName );
            if( stat( path, &s_buf ) == -1 ) {
                fprintf( stderr, "File \"%s\" does not exist 1\n\n", FileName );
                exit( -1 );
            }
            object_defined = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-sofas", 6) == 0) ) {
            sofas = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-trackball", 8) == 0) ) {
            track = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-f", 2) == 0) || 
            (strcmp(argv[i], "-file") == 0) ) {
            if (++i >= argc) usage( argv[0] );
            file_defined = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-h", 2) == 0) || 
            (strcmp(argv[i], "-help") == 0) ) {
            usage( argv[0] );
            continue;
        }
        else if( (strncmp(argv[i], "-na", 3) == 0) ||
            (strcmp(argv[i], "-naxis") == 0) ) {
            no_axis = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-nbg", 4) == 0) ||
            (strcmp(argv[i], "-nbackground") == 0) ) {
            no_background = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-nda", 4) == 0) ||
            (strcmp(argv[i], "-ndarea") == 0) ) {
            no_darea = 0;
            continue;
        }
        else if( (strncmp(argv[i], "-nm", 3) == 0) ||
            (strcmp(argv[i], "-nmove") == 0) ) {
            no_move = 1;
            continue;
        }
        else if( (strncmp(argv[i], "-v", 2) == 0) ||
            (strcmp(argv[i], "-verbose") == 0) ) {
            verbose = 1;
            continue;
        }
        else
            usage( argv[0] );
    }
}


void usage( program_name )
char *program_name;
{
   fprintf( stderr, "Usage: %s \n\
       -display display \n\
       -h[elp] \n\
       -sofas (use SetOfFillAreaSets instead of FillAreaWithData) \n\
       -trackball (rotations) \n\
       -nbg[nbackground] (does not put up the background) \n\
       -nda[rea] (draws to a xmDrawingArea instead of a window) \n\
       -nm[ove] (does not allow windows to move) \n\
       -d[rawable] buffer | pixmap | window \n\
       -v[erbose] \n\
       -o[bject] object \n", program_name );
   exit(-1);
}


void
AppMainLoop()
{
    extern XtActionProc CanvasButtonPress(), CanvasButtonRelease(),
        CanvasB1Motion(), CanvasB2Motion(), CanvasB3Motion(),
        CanvasExpose(), CanvasConfigure(), CanvasResize(), CanvasReset();
    extern XtActionProc quitOFFObject();

    extern Display *display;
    extern Window   window;
    XEvent          event;
    Widget          w;

    for( ;; ) {
/*
        if( XtAppPending( app_context ) ) {
*/
            XtAppNextEvent( app_context, &event );
            if( event.xany.window == window ) {

            switch( event.type ) {

            case Expose:
                while( XCheckTypedWindowEvent( display, window, Expose,
                    &event ));

                CanvasExpose( w, &event );
                break;

            case ConfigureNotify:
                break;

            case ButtonPress:
                CanvasButtonPress( w, &event );
                break;

            case ButtonRelease:
                CanvasButtonRelease( w, &event );
                break;

            case MotionNotify:
                while( XCheckTypedWindowEvent(display, window, MotionNotify,
                    &event ));

                     if( event.xmotion.state == Button1MotionMask )
                    CanvasB1Motion( w, &event );
                else if( event.xmotion.state == Button2MotionMask )
                    CanvasB2Motion( w, &event );
                else if( event.xmotion.state == Button3MotionMask )
                    CanvasB3Motion( w, &event );

                break;

            case KeyPress: {
                KeySym keysym;
                char buf[42];
                XComposeStatus cs;
                int length = XLookupString( &event.xkey, buf, 42, &keysym,&cs );
                switch( keysym ) {
                case XK_q:
                case XK_Q:
                    quitOFFObject( w );
                    break;
                case XK_r:
                case XK_R:
                    CanvasReset();
                    break;
                case XK_Alt_L:
                case XK_Control_L:
                case XK_Meta_L:
                case XK_Shift_L:
                    left_key_press = 1;
                    break;
                default:
                    break;
                }
                break;
            }
            case KeyRelease: {
                KeySym keysym;
                char buf[42];
                XComposeStatus cs;
                int length = XLookupString( &event.xkey, buf, 42, &keysym,&cs );
                switch( keysym ) {
                case XK_Alt_L:
                case XK_Control_L:
                case XK_Meta_L:
                case XK_Shift_L:
                    left_key_press = 0;
                    break;
                default:
                    break;
                }
                break;
            }
            case MapNotify:
            case PropertyNotify:
                break;
            default:
                if( verbose )
                printEvent( &event );
                break;
            }
            }
            else
                XtDispatchEvent( &event );
/*
        }
*/
    }
}
